<div class="pagination">
    <span class="step-links">
        {% if page_obj.has_previous %}
        <a href="?page=1">&laquo; Première page</a>
        <a href="?page={{ page_obj.previous_page_number }}">Page précédente</a>
        {% endif %}

        <span class="current">
            Page {{ page_obj.number }} / {{ page_obj.paginator.num_pages }}
        </span>

        {% if page_obj.has_next %}
        <a href="?page={{ page_obj.next_page_number }}">Page suivante</a>
        <a href="?page={{ page_obj.paginator.num_pages }}">Dernière page</a>
        {% endif %}
    </span>
</div>

